/*
 *  Copyright (C) 2003 2004
 *  Stefano Tonetta <stonetta@dit.unitn.it>
 *  Roberto Sebastiani <rseba@dit.unitn.it>
 *
*/
#include "main.h"
#include "automa.h"
#include "form.h"
#include "cover.h"
#include "struct.h"
#include "prop.h"
#include "utils.h"
#include "modella.y.tab.h" 


GBA* new_GBA(form* f){
  int i,j,k;
  GBA* b=(GBA*)mymalloc(sizeof(GBA));
  int states_number=0;
  int initial_states=0;
  int real_fair_number=0;
  char** F;
  cover* c=new_cover(f);
  queue* q=new_queue();
  list* tmp=c;
  list* states=NULL;
  while(tmp!=NULL){
    xstate* xs=tmp->el;
    list *tmp2=states;
    while(tmp2!=NULL){
      xstate* xs2=tmp2->el;
      if(xs->s==xs2->s){
	tmp2=NULL;
	xs=NULL;
      } else 
	tmp2=tmp2->next;
    }
    if(xs!=NULL){
      states=new_list(xs,states);
      enqueue(xs,q);
      states_number++;
      initial_states++;
    }
    tmp=tmp->next;
  }
  while(!empty_queue(q)){
    xstate* xs;
    xs=dequeue(q);
    c=new_cover(xs->s->right->left);
    tmp=c;
    while(tmp!=NULL){
      xstate* xs=tmp->el;
      list *tmp2=states;
      while(tmp2!=NULL){
	xstate* xs2=tmp2->el;
	if(xs->s==xs2->s){
	  tmp2=NULL;
	  xs=NULL;
	} else 
	  tmp2=tmp2->next;
      }
      if(xs!=NULL){
	states=new_list(xs,states);
	enqueue(xs,q);
	states_number++;
      }
      tmp=tmp->next;
    }
  }
  
  b->states_number=states_number;
  b->prop_number=prop_number;
  b->initial_states=initial_states;
  
  b->Q=(xstate**)mymalloc(states_number*sizeof(xstate*));
  i=states_number;
  tmp=states;
  while(tmp!=NULL){
    list* tmp2=tmp;
    b->Q[--i]=tmp->el;
    tmp=tmp->next;
    free(tmp2);
  }
	
  b->R=(char**)mymalloc(states_number*sizeof(char*));
  for(i=0;i<states_number;i++){
    b->R[i]=(char*)mymalloc(states_number*sizeof(char));
    for(j=0;j<states_number;j++)
      b->R[i][j]=0;
    c=new_cover(b->Q[i]->s->right->left);
    tmp=c;
    while(tmp!=NULL){
      xstate* xs=tmp->el;
      for(j=0;j<states_number && b->Q[j]->s!=xs->s;j++);
      b->R[i][j]=1;
      tmp=tmp->next;
    }
  }

  F=(char**)mymalloc(fair_number*sizeof(char*));
  for(i=0;i<fair_number;i++){
    int full=1;
    F[i]=(char*)mymalloc(states_number*sizeof(char));
    for(j=0;j<states_number;j++)
      if(xsinF(b->Q[j],i))
	F[i][j]=1;
      else {
	F[i][j]=0;
	full=0;
      }
    if(full){
      free(F[i]);
      F[i]=NULL;
    } 
    for(j=0;j<i && F[i]!=NULL;j++)
      if(F[j]!=NULL){
	int geq=1;
	for(k=0;k<states_number && geq;k++)//check if F_i is supset of F_j
	  if(F[i][k]<F[j][k])
	    geq=0;
	if(geq){
	  free(F[i]);
	  F[i]=NULL;
	} else {
	  int leq=1;
	  for(k=0;k<states_number && leq;k++)//check if F_j is supset of F_i
	    if(F[i][k]>F[j][k])
	      leq=0;
	  if(leq){
	    free(F[j]);
	    F[j]=NULL;
	    real_fair_number--;
	  }
	}
      }
    if(F[i]!=NULL)
      real_fair_number++;
  }
  j=0;
  b->fair_number=real_fair_number;
  b->F=(char**)mymalloc(real_fair_number*sizeof(char*));
  for(i=0;i<fair_number;i++)
    if(F[i]!=NULL)
      b->F[j++]=F[i];
  free(F);
  return b;


}



void print_GBAw(GBA* b,FILE* output){
  int i,j;
  fprintf(output,"States\n");
  for(i=0;i<b->states_number;i++){
    fprintf(output,"n%d: {",i);
    print_form_wring(b->Q[i]->s,stdout);
    fprintf(output,"} label: {");
    print_form_wring(b->Q[i]->s->left,stdout);
    fprintf(output,"}\n");
  }
  fprintf(output,"Arcs\n");
  for(i=0;i<b->states_number;i++){
    int firstt=1;
    if(i<b->initial_states)
      fprintf(output,"-> ");
    else
      fprintf(output,"   ");
    fprintf(output,"n%d -> {",i);
    for(j=0;j<b->states_number;j++)
      if(b->R[i][j]){
	if(firstt){
	  fprintf(output,"n%d",j);
	  firstt=0;
	} else
	  fprintf(output,",n%d",j);
      }
    fprintf(output,"}\n");
  }
  fprintf(output,"Fair Sets\n");
  for(i=0;i<b->fair_number;i++){
    int firsts=1;
    fprintf(output,"{");
    for(j=0;j<b->states_number;j++)
      if(b->F[i][j]){
	if(firsts){
	  fprintf(output,"n%d",j);
	  firsts=0;
	} else
	  fprintf(output,",n%d",j);
      }
    fprintf(output,"}\n");
  }
  fprintf(output,"End\n");
}


void print_GBA2(GBA* b,FILE* output){
  int i,j;
  fprintf(output,"GBA:\n");
  fprintf(output,"%d\n",b->states_number);
  fprintf(output,"%d\n",b->prop_number);
  fprintf(output,"%d\n",b->initial_states);
  fprintf(output,"%d\n",b->fair_number);
  for(i=0;i<b->states_number;i++){
    print_form_prefix(b->Q[i]->s,stdout);
    fprintf(output,"\n");
  }
  for(i=0;i<b->states_number;i++){
    for(j=0;j<b->states_number;j++)
      fprintf(output,"%d ",b->R[i][j]);
    fprintf(output,"\n");
  }
  for(i=0;i<b->fair_number;i++){
    for(j=0;j<b->states_number;j++)
      fprintf(output,"%d ",b->F[i][j]);
    fprintf(output,"\n");
  }
}


void print_GBA(GBA* b,FILE* output){
  int i,j;
  fprintf(output,"%d %d ",b->states_number+1,b->fair_number);
  fprintf(output,"0 1 -1 ");
  for(i=0;i<b->initial_states;i++){
    fprintf(output,"%d ",i+1);
    print_form_prefix(b->Q[i]->s->left,output);
  }
  fprintf(output,"-1 ");
  for(i=0;i<b->states_number;i++){
    fprintf(output,"%d 0 ",i+1);
    for(j=0;j<b->fair_number;j++)
      if(b->F[j][i])
	fprintf(output,"%d ",j);
    fprintf(output,"-1 ");
    for(j=0;j<b->states_number;j++)
      if(b->R[i][j]){
	fprintf(output,"%d ",j+1);
	print_form_prefix(b->Q[j]->s->left,output);
      }
    fprintf(output,"-1 ");
  
  }
}


BA* GBAtoBA(GBA* b1){
  int i,j;
  if(b1->fair_number<2){
    BA* b2=(BA*)mymalloc(sizeof(BA));
    b2->states_number=b1->states_number;
    b2->initial_states=b1->initial_states;
    b2->prop_number=b1->prop_number;
    b2->Q=b1->Q;
    b2->R=b1->R;
    if(b1->fair_number==1)
      b2->F=b1->F[0];
    else {
      b2->F=(char*)mymalloc(b2->states_number*sizeof(char));
      for(i=0;i<b2->states_number;i++)
	b2->F[i]=1;
    }
    return b2;
  } else {
    int gfair;
    int count=0;
    int greatest=0;
    BA* b2=(BA*)mymalloc(sizeof(BA));
    b2->states_number=b1->states_number*b1->fair_number;
    b2->initial_states=b1->initial_states;
    b2->prop_number=b1->prop_number;
    b2->Q=(xstate**)mymalloc(b2->states_number*sizeof(xstate*));
    b2->R=(char**)mymalloc(b2->states_number*sizeof(char*));
    for(i=0;i<b2->states_number;i++)
      b2->R[i]=(char*)mymalloc(b2->states_number*sizeof(char));
    b2->F=(char*)mymalloc(b2->states_number*sizeof(char));
    for(i=0;i<b2->states_number;i++)
      b2->Q[i]=b1->Q[i%b1->states_number];
    for(i=0;i<b2->states_number;i++)
      for(j=0;j<b2->states_number;j++)
	if(i/b1->states_number==j/b1->states_number){
	  if(b1->F[i/b1->states_number][i%b1->states_number])
	    b2->R[i][j]=0;
	  else
	    b2->R[i][j]=b1->R[i%b1->states_number][j%b1->states_number];
	} else if((i/b1->states_number+1)%b1->fair_number==j/b1->states_number){
	  if(b1->F[i/b1->states_number][i%b1->states_number])
	    b2->R[i][j]=b1->R[i%b1->states_number][j%b1->states_number];
	  else
	    b2->R[i][j]=0;
	} else
	  b2->R[i][j]=0;
    

    
    for(i=0;i<b1->fair_number;i++){
      for(j=0;j<b1->states_number;j++)
	if(b1->F[i][j])
	  count++;
      if(count>greatest){
	greatest=count;
	gfair=i;
      }
    }
      

    for(i=0;i<b2->states_number;i++)
      if(i/b1->states_number==gfair)
	b2->F[i]=b1->F[gfair][i%b1->states_number];
      else
	b2->F[i]=0;
    return b2;
  }
}

int compute_reachable_BA(BA* b,char* reach){
  int i;
  int states_number=0;
  queue* q=new_queue();
  for(i=0;i<b->states_number;i++)
    reach[i]=0;
  for(i=0;i<b->initial_states;i++){
    int* s=(int*)mymalloc(sizeof(int));
    *s=i;
    reach[i]=1;
    enqueue(s,q);
    states_number++;
  }
  while(!empty_queue(q)){
    int *s=(int*)dequeue(q);
    for(i=0;i<b->states_number;i++){
      if(b->R[*s][i] && !reach[i]){
	int* s2=(int*)mymalloc(sizeof(int));
	*s2=i;
	reach[i]=1;
	enqueue(s2,q);
	states_number++;
      }
    }
  }
  free(q);
  return states_number;
}

void print_BA2(BA* b,FILE* output){
  int i,j;
  fprintf(output,"BA\n");
  fprintf(output,"%d\n",b->states_number);
  fprintf(output,"%d\n",b->prop_number);
  fprintf(output,"%d\n",b->initial_states);
  for(i=0;i<b->states_number;i++){
    print_form_prefix(b->Q[i]->s,stdout);
    fprintf(output,"\n");
  }
  for(i=0;i<b->states_number;i++){
    for(j=0;j<b->states_number;j++)
      fprintf(output,"%d ",b->R[i][j]);
    fprintf(output,"\n");
  }
  for(j=0;j<b->states_number;j++)
    fprintf(output,"%d ",b->F[j]);
  fprintf(output,"\n");
}

void print_BA(BA* b,FILE* output){
  int i,j;
  char* reach=(char*)mymalloc(b->states_number*sizeof(char));
  int states_number=compute_reachable_BA(b,reach);
  fprintf(output,"%d 1 ",states_number+1);
  fprintf(output,"0 1 -1 ");
  for(i=0;i<b->initial_states;i++){
    fprintf(output,"%d ",i+1);
    print_form_prefix(b->Q[i]->s->left,output);
  }
  fprintf(output,"-1 ");
  for(i=0;i<b->states_number;i++)
    if(reach[i]){
      fprintf(output,"%d 0 ",i+1);
      if(b->F[i])
	fprintf(output,"%d ",1);
      fprintf(output,"-1 ");
      for(j=0;j<b->states_number;j++)
	if(b->R[i][j]){
	  fprintf(output,"%d ",j+1);
	  print_form_prefix(b->Q[j]->s->left,output);
	}
      fprintf(output,"-1 ");
      
    }
}



void BAtodot(BA* b){ 
  int i,j;
  FILE* dotfile;
  dotfile=fopen("automa.dot","w");
  if(dotfile!=NULL){
    fprintf(dotfile,"digraph modella {\n");
    for(i=0;i<b->initial_states;i++){
      fprintf(dotfile,"init -> n%d;\n",i);
    }
    for(i=0;i<b->states_number;i++){
      for(j=0;j<b->states_number;j++)
	if(b->R[i][j]){
	  fprintf(dotfile,"n%d -> n%d;\n",i,j);
	}
    }
    fprintf(dotfile,"init [shape=plaintext,label=\"\"];\n");
    for(i=0;i<b->states_number;i++)
      if(b->F[i]){
	fprintf(dotfile,"n%d [shape=box,label=\"",i);
	print_form_prefix(b->Q[i]->s,dotfile);
	fprintf(dotfile,"\"];\n");
      } else {
	fprintf(dotfile,"n%d [shape=ellipse,label=\"",i);
	print_form_prefix(b->Q[i]->s,dotfile);
	fprintf(dotfile,"\"];\n");
      }
    fputs("}\n",dotfile);
    fclose(dotfile);
    if (!system("dot -Tps automa.dot -o automa.ps"))
      system("ghostview automa.ps");
    else fatal_error("BAtodot: impossibile generare il file .ps");
  } else fatal_error("BAtodot: impossibile aprire il file");
  
}




void GBAtodot(GBA* b){ 
  int i,j;
  FILE* dotfile;
  dotfile=fopen("automa.dot","w");
  if(dotfile!=NULL){
    fprintf(dotfile,"digraph modella {\n");
    for(i=0;i<b->initial_states;i++){
      fprintf(dotfile,"init -> n%d;\n",i);
    }
    for(i=0;i<b->states_number;i++){
      for(j=0;j<b->states_number;j++)
	if(b->R[i][j]){
	  fprintf(dotfile,"n%d -> n%d;\n",i,j);
	}
    }
    fprintf(dotfile,"init [shape=plaintext,label=\"\"];\n");
    for(i=0;i<b->states_number;i++){
      fprintf(dotfile,"n%d [shape=ellipse,label=\"",i);
      print_form_prefix(b->Q[i]->s,dotfile);
      for(j=0;j<b->fair_number;j++)
	if(b->F[j][i])
	  fprintf(dotfile," (%d)",j);
      fprintf(dotfile,"\"];\n");
    }
    fputs("}\n",dotfile);
    fclose(dotfile);
    if (!system("dot -Tps automa.dot -o automa.ps"))
      system("ghostview automa.ps");
    else fatal_error("impossibile generare il file .ps");
  } else fatal_error("impossibile aprire il file");
  
}


int TarjanDFS(int v,int count,char *visited,char* root,char* comp,char* lang_not_empty,LBA* b){
  int i;
  int old_count=count;
  int reach_fairSCC=0;
  visited[v]=1;
  root[v]=count++;
  comp[v]=0;
  if(b->F[v])
    lang_not_empty[v]=1;
  else
    lang_not_empty[v]=0;
  for(i=0;i<b->trans_number;i++)
    if(b->R[i]->source==v){
      int w=b->R[i]->dest;
      if(visited[w]==0)
	count=TarjanDFS(w,count,visited,root,comp,lang_not_empty,b);
      if(comp[w]==0)
	if(root[w]<root[v])
	  root[v]=root[w];
      if(lang_not_empty[w]==1)
	reach_fairSCC=1;
      if(comp[w]==0 && lang_not_empty[v]==1)
	reach_fairSCC=1;
    }
  if(root[v]==old_count){
    for(i=0;i<b->states_number;i++)
      if(root[i]==root[v]){
	comp[i]=1;
	lang_not_empty[i]=reach_fairSCC;
      }
  } else
    lang_not_empty[v]=reach_fairSCC;
  return count;
}


LBA* prune(LBA* b1){
  char *visited=(char*)mymalloc(b1->states_number*sizeof(char));
  char *root=(char*)mymalloc(b1->states_number*sizeof(char));
  char *comp=(char*)mymalloc(b1->states_number*sizeof(char));
  char *lang_not_empty=(char*)mymalloc(b1->states_number*sizeof(char));
  LBA* b2=(LBA*)mymalloc(sizeof(LBA));
  int states_number=0;
  int trans_number=0;
  int* id=(int*)mymalloc(b1->states_number*sizeof(int));
  int i,j;
  for(i=0;i<b1->states_number;i++){
    visited[i]=0;
    root[i]=-1;
  }
  TarjanDFS(0,0,visited,root,comp,lang_not_empty,b1);
  for(i=0;i<b1->states_number;i++)
    if(visited[i] && lang_not_empty[i])
      id[i]=states_number++;
  for(i=0;i<b1->trans_number;i++)
    if(visited[b1->R[i]->source] && visited[b1->R[i]->dest] && lang_not_empty[b1->R[i]->source] && lang_not_empty[b1->R[i]->dest])
      trans_number++;
  b2->states_number=states_number;
  b2->trans_number=trans_number;
  b2->prop_number=b1->prop_number;
  b2->R=(transition**)mymalloc(b2->trans_number*sizeof(transition*));
  for(i=0;i<b2->trans_number;i++)
    b2->R[i]=(transition*)mymalloc(sizeof(transition));
  for(i=0,j=0;i<b1->trans_number;i++)
    if(visited[b1->R[i]->source] && visited[b1->R[i]->dest] && lang_not_empty[b1->R[i]->source] && lang_not_empty[b1->R[i]->dest]){
      b2->R[j]->source=id[b1->R[i]->source];
      b2->R[j]->dest=id[b1->R[i]->dest];
      b2->R[j]->label=b1->R[i]->label;
      j++;
    }
  b2->F=(char*)mymalloc(b2->states_number*sizeof(char));
  for(i=0;i<b1->states_number;i++)
    if(lang_not_empty[i]){
      int triv=1;
      for(j=0;triv && j<b1->trans_number;j++)
	if(b1->R[j]->source==i && root[b1->R[j]->dest]==root[i])
	  triv=0;
      if(triv)
	b2->F[id[i]]=0;
      else
	b2->F[id[i]]=b1->F[i];
    }
  free(visited);
  free(root);
  free(comp);
  free(lang_not_empty);
  free(id);
  for(i=0;i<b1->trans_number;i++)
    free(b1->R[i]);
  free(b1->R);
  free(b1->F);
  free(b1);
  return b2;
}


int compute_reachable_LBA(LBA* b,char* reach){
  int i;
  int* s;
  int states_number=0;
  queue* q=new_queue();
  for(i=0;i<b->states_number;i++)
    reach[i]=0;
  s=(int*)mymalloc(sizeof(int));
  *s=0;
  reach[0]=1;
  enqueue(s,q);
  states_number++;
  while(!empty_queue(q)){
    int *s=(int*)dequeue(q);
    for(i=0;i<b->trans_number;i++){
      if(b->R[i]->source==*s && !reach[b->R[i]->dest]){
	int* s2=(int*)mymalloc(sizeof(int));
	*s2=b->R[i]->dest;
	reach[b->R[i]->dest]=1;
	enqueue(s2,q);
	states_number++;
      }
    }
    free(s);
  }
  //free(q);
  return states_number;
}

void print_LBA2(LBA* b,FILE* output){
  int i,j;
  fprintf(output,"LBA:\n");
  fprintf(output,"%d\n",b->states_number);
  fprintf(output,"%d\n",b->trans_number);
  fprintf(output,"%d\n",b->prop_number);
  for(i=0;i<b->trans_number;i++){
    print_form_prefix(b->R[i]->label,stdout);
    fprintf(output,"%d %d ",b->R[i]->source,b->R[i]->dest);
    fprintf(output,"\n");
  }
  for(j=0;j<b->states_number;j++)
    fprintf(output,"%d ",b->F[j]);
  fprintf(output,"\n");
}

void print_LBA(LBA* b,FILE* output){
  int i,j;
  //char* reach=(char*)mymalloc(b->states_number*sizeof(char));
  //int states_number=compute_reachable_LBA(b,reach);
  fprintf(output,"%d 1 ",b->states_number);
  for(i=0;i<b->states_number;i++){
    //if(reach[i]){
      fprintf(output,"%d ",i);
      if(i==0)
	fprintf(output,"1 ");
      else
	fprintf(output,"0 ");
      if(b->F[i])
	fprintf(output,"1 ");
      fprintf(output,"-1 ");
      for(j=0;j<b->trans_number;j++)
	if(b->R[j]->source==i){
	  fprintf(output,"%d ",b->R[j]->dest);
	  print_form_prefix(b->R[j]->label,output);
	  fprintf(output,"\n"); // required by lbtt 1.1.x
	}
      fprintf(output,"-1 ");
      
    }
}

void print_LBAs(LBA* b,FILE* output){
  int i,j;
  int skip;
  if(b->states_number==0)
    return;
  fprintf(output,"never{\n");
  for(i=0;i<b->states_number;i++){
    if(b->F[i])
      fprintf(output,"accept_");
    fprintf(output,"T%d",i);
    if(i==0)
      fprintf(output,"_init");
    fprintf(output,":\n");
    if(b->F[i])
      skip=0;
    else 
      skip=-1;
    for(j=0;j<b->trans_number && skip>=0;j++)
      if(b->R[j]->source==i){
	if(b->R[j]->dest==i && b->R[j]->label->type==TRUE)
	  skip=1;
	else
	  skip=-1;
      }
    if(skip==1)
      fprintf(output,"\tskip;\n");
    else {
      fprintf(output,"\tif\n");
      for(j=0;j<b->trans_number;j++)
	if(b->R[j]->source==i){
	  fprintf(output,"\t:: ");
	  print_form_spin(b->R[j]->label,output);
	  fprintf(output," -> goto ");
	  if(b->F[b->R[j]->dest])
	    fprintf(output,"accept_");
	  fprintf(output,"T%d",b->R[j]->dest);
	  if(b->R[j]->dest==0)
	    fprintf(output,"_init");
	  fprintf(output,"\n");
	}
      fprintf(output,"\tfi;\n");
    }
    
  }
  fprintf(output,"}\n");
  
}




void print_LBAs_l2s(LBA* b,FILE* output){
  int i,j,k;
  int skip;
  if(b->states_number==0)
    return;
  fprintf(output,"never{\n");
  for(i=0;i<b->states_number;i++){
    fprintf(output,"T%d",i);
    if(i==0)
      fprintf(output,"_init");
    fprintf(output,":\n");
    fprintf(output,"\tif\n");
    for(j=0;j<b->trans_number;j++)
      if(b->R[j]->source==i){
	fprintf(output,"\t:: ");
	print_form_spin(b->R[j]->label,output);
	fprintf(output," -> goto ");
	fprintf(output,"T%d",b->R[j]->dest);
	if(b->R[j]->dest==0)
	  fprintf(output,"_init");
	fprintf(output,"\n");
	if(b->F[i]){
	  fprintf(output,"\t:: ((");
	  print_form_spin(b->R[j]->label,output);
	  fprintf(output,") && __l2s_save__ && !__l2s_saved__) -> goto ");
	  fprintf(output,"T%d_%d",b->R[j]->dest,i);
	  fprintf(output,"\n");
	}
      }
    fprintf(output,"\tfi;\n");
  }
  for(k=0;k<b->states_number;k++)
    if(b->F[k]){
      for(i=0;i<b->states_number;i++){
	fprintf(output,"T%d_%d",i,k);
	fprintf(output,":\n");
	fprintf(output,"\tif\n");
	for(j=0;j<b->trans_number;j++)
	  if(b->R[j]->source==i){
	    fprintf(output,"\t:: ((");
	    print_form_spin(b->R[j]->label,output);
	    fprintf(output,") && __l2s_saved__) -> goto ");
	    fprintf(output,"T%d_%d",b->R[j]->dest,k);
	    fprintf(output,"\n");
	    if(b->R[j]->dest==k){
	      fprintf(output,"\t:: ((");
	      print_form_spin(b->R[j]->label,output);
	      fprintf(output,") && __l2s_looped__) -> goto ");
	      fprintf(output,"accept_T%d",b->R[j]->dest);
	      fprintf(output,"\n");
	    }
	  }
	fprintf(output,"\tfi;\n");
      }
    }
  for(k=0;k<b->states_number;k++)
    if(b->F[k]){
      fprintf(output,"accept_T%d",k);
      fprintf(output,":\n");
      fprintf(output,"\tskip;\n");
    }
  fprintf(output,"}\n");
  
}



void print_LBAn(LBA* b,FILE* output){
  int i;
  fprintf(output,"<buechi type=\"BA\">\n");
  for(i=0;i<b->states_number;i++){
    fprintf(output,"<vertex id=\"v%d\"",i);
    if(i==0)
      fprintf(output," initial=\"1\"/>\n");
    else
      fprintf(output,"/>\n");
  }
  for(i=0;i<b->trans_number;i++){
    fprintf(output,"<edge from=\"v%d\" to=\"v%d\" label=\"",
	      b->R[i]->source,b->R[i]->dest);
    print_form_nusmv(b->R[i]->label,output);
    fprintf(output,"\"/>\n");
  }
  fprintf(output,"<fairset set=\" ");
  for(i=0;i<b->states_number;i++)
    if(b->F[i])
      fprintf(output,"v%d ",i);
  fprintf(output,"\"/>\n");
  fprintf(output,"</buechi>");
}






int TarjanDFS_weak(int v,int count,int *visited,int* root,char* comp,GBA* b){
  int i,w,trivial=1;
  visited[v]=count++;
  root[v]=v;
  comp[v]=0;
  for(w=0;w<b->states_number;w++)
    if(b->R[v][w]){
      if(visited[w]==0)
	count=TarjanDFS_weak(w,count,visited,root,comp,b);
      if(comp[w]==0){
	trivial=0;
	if(visited[w]<visited[v]){
	  visited[v]=visited[w];
	  root[v]=root[w];
	}
      }
    }
  if(trivial)
    comp[v]=-1;
  else {
    if(root[v]==v)
      for(i=0;i<b->states_number;i++)
	if(root[i]==root[v])
	  comp[i]=1;
  }
  return count;
}



static char* check_weak(GBA* b){
  int *visited=mymalloc(b->states_number*sizeof(int));
  int *root=mymalloc(b->states_number*sizeof(int));
  char *comp=mymalloc(b->states_number*sizeof(char));
  char *fair=mymalloc(b->states_number*sizeof(char));
  char *tmp=mymalloc(b->states_number*sizeof(char));
  int i,j,v;
  int count=1;
  for(i=0;i<b->states_number;i++)
    visited[i]=0;
  for(i=0;i<b->initial_states;i++){
    if(visited[i]==0)
      count=TarjanDFS_weak(i,count,visited,root,comp,b);
  }
  for(i=0;i<b->fair_number;i++){
    for(j=0;j<b->states_number;j++)
      if(b->F[i][j])
	fair[j]=1;
      else
	fair[j]=0;
    for(j=0;j<b->states_number;j++)
      if(fair[j]!=fair[root[j]]){
	myfree(visited);
	myfree(root);
	myfree(comp);
	myfree(fair);
	myfree(tmp);
	return 0;
      }
  }
  for(i=0;i<b->states_number;i++)
    tmp[i]=1;
  for(i=0;i<b->fair_number;i++){
    for(j=0;j<b->states_number;j++)
      if(!b->F[i][j] || comp[j]==-1)
	tmp[j]=0;
  }
  return tmp;
}
    


void print_GBAn(GBA* b,FILE* output){
  int i,j;
  char *fair=NULL;
  fprintf(output,"<buechi type=\"LBA\">\n");
  for(i=0;i<b->states_number;i++){
    fprintf(output,"<vertex id=\"v%d\" label=\"",i);
    print_form_nusmv(b->Q[i]->s->left,output);
    if(i<b->initial_states)
      fprintf(output,"\" initial=\"1\"/>\n");
    else
      fprintf(output,"\"/>\n");
  }
  for(i=0;i<b->states_number;i++)
    for(j=0;j<b->states_number;j++)
      if(b->R[i][j])
	fprintf(output,"<edge from=\"v%d\" to=\"v%d\"/>\n",i,j);
  if(weak)
    fair=check_weak(b);
  if(fair){
    fprintf(output,"<fairset set=\" ");
    for(j=0;j<b->states_number;j++)
      if(fair[j])
	fprintf(output,"v%d ",j);
    fprintf(output,"\"/>\n");
  } else {
    for(i=0;i<b->fair_number;i++){
      fprintf(output,"<fairset set=\" ");
      for(j=0;j<b->states_number;j++)
	if(b->F[i][j])
	  fprintf(output,"v%d ",j);
      fprintf(output,"\"/>\n");
    }
  }
  fprintf(output,"</buechi>");

}


void print_GBAm(GBA* b,FILE* output){
  int i,j;
  char *fair=NULL;
  fprintf(output,"MODULE ltl_spec_0\n");
  fprintf(output,"VAR\n");
  fprintf(output,"   LTL_0_SPEC_state : 0..%d;\n",b->states_number);
  fprintf(output,"INIT\n");
  if(b->initial_states==0)
    fprintf(output,"   FALSE\n");
  else 
    fprintf(output,"   (LTL_0_SPEC_state=0)\n");
  for(i=1;i<b->initial_states;i++)
    fprintf(output,"   |(LTL_0_SPEC_state=%d)\n",i);
  for(i=0;i<b->states_number;i++){
    int first=1;
    fprintf(output,"TRANS\n");
    fprintf(output,"   (LTL_0_SPEC_state=%d)->next(LTL_0_SPEC_state in {",i);
    for(j=0;j<b->states_number;j++)
      if(b->R[i][j]){
	if(first){
	  first=0;
	  fprintf(output,"%d",j);
	} else 
	  fprintf(output,",%d",j);
      }
    fprintf(output,"})\n");
  }
  for(i=0;i<b->states_number;i++){
    fprintf(output,"INVAR\n");
    fprintf(output,"   (LTL_0_SPEC_state=%d)->",i);
    print_form_nusmv(b->Q[i]->s->left,output);
    fprintf(output,"\n");
  }
  if(weak)
    fair=check_weak(b);
  if(fair){
    int first=1;
    fprintf(output,"DEFINE\n");
    fprintf(output,"   LTL_0_SPECF_bad := ");
    for(j=0;j<b->states_number;j++)
      if(fair[j]){
	if(first){
	  first=0;
	  fprintf(output,"(LTL_0_SPEC_state=%d)",j);
	} else 
	  fprintf(output,"|(LTL_0_SPEC_state=%d)",j);
      }
    if(first)
      fprintf(output,"FALSE;\n");
    else
      fprintf(output,";\n");
  } else {
    for(i=0;i<b->fair_number;i++){
      int first=1;
      fprintf(output,"FAIRNESS\n");
      for(j=0;j<b->states_number;j++)
	if(b->F[i][j]){
	  if(first){
	    first=0;
	    fprintf(output,"(LTL_0_SPEC_state=%d)",j);
	  } else 
	    fprintf(output,"|(LTL_0_SPEC_state=%d)",j);
	}
      if(first)
	fprintf(output,"FALSE;\n");
      else
	fprintf(output,"\n");
    }
  }
	
  
}



void LBAtodot(LBA* b){ 
  int i;
  FILE* dotfile;
  dotfile=fopen("automa.dot","w");
  if(dotfile!=NULL){
    fprintf(dotfile,"digraph modella {\n");
    fprintf(dotfile,"init -> n0;\n");
    for(i=0;i<b->trans_number;i++){
      fprintf(dotfile,"n%d -> n%d [label=\"",b->R[i]->source,b->R[i]->dest);
	  print_form_prefix(b->R[i]->label,dotfile);
	  fprintf(dotfile,"\"];\n");
    }
    fprintf(dotfile,"init [shape=plaintext,label=\"\"];\n");
    for(i=0;i<b->states_number;i++){
      fprintf(dotfile,"n%d [shape=ellipse,label=\"n%d",i,i);
      if(b->F[i])
	fprintf(dotfile," (1)");
      fprintf(dotfile,"\"];\n");
    }
    fputs("}\n",dotfile);
    fclose(dotfile);
    if (!system("dot -Tps automa.dot -o automa.ps"))
      system("ghostview automa.ps");
    else fatal_error("impossibile generare il file .ps");
  } else fatal_error("impossibile aprire il file");
  
}



LBA* BAtoLBA(BA* b1){
  int i,j,k;
  int trans_number=b1->initial_states;
  LBA* b2=(LBA*)mymalloc(sizeof(LBA));
  for(i=0;i<b1->states_number;i++)
    for(j=0;j<b1->states_number;j++)
      if(b1->R[i][j])
	trans_number++;
  b2->states_number=b1->states_number+1;
  b2->trans_number=trans_number+b1->initial_states;
  b2->prop_number=b1->prop_number;
  b2->R=(transition**)mymalloc(b2->trans_number*sizeof(transition*));
  for(i=0;i<b2->trans_number;i++)
    b2->R[i]=(transition*)mymalloc(sizeof(transition));
  b2->F=(char*)mymalloc(b2->states_number*sizeof(char));
  trans_number=0;
  for(i=0;i<b1->initial_states;i++){
    b2->R[trans_number]->source=0;
    b2->R[trans_number]->dest=i+1;
    b2->R[trans_number]->label=b1->Q[i]->s->left;
    trans_number++;
  }
  for(i=0;i<b1->states_number;i++)
    for(j=0;j<b1->states_number;j++)
      if(b1->R[i][j]){
	for(k=0;k<trans_number && (b2->R[k]->source!=i+1 || b2->R[k]->dest!=j+1 || b2->R[k]->label!=b1->Q[j]->s->left);k++);
	if(k==trans_number){	  
	  b2->R[trans_number]->source=i+1;
	  b2->R[trans_number]->dest=j+1;
	  b2->R[trans_number]->label=b1->Q[j]->s->left;
	  trans_number++;
	}
      }
  b2->trans_number=trans_number;
  b2->F[0]=1;
  for(j=1;j<b2->states_number;j++)
    b2->F[j]=b1->F[j-1];	  
  if(verbosity>1)
    print_LBA2(b2,stdout);
  return b2;
}

