/**CFile***********************************************************************

  FileName    [cover.c]

  Package     []

  Synopsis    [Cover computation.]

  Description [This file provides the function to compute the cover of
  an LTL formula.]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.9]

******************************************************************************/

#include <stdio.h>
#include "struct.h"
#include "cover.h"
#include "form.h"
#include "prop.h"
#include "main.h"
#include "modella.y.tab.h" 

list *covered_forms=NULL;


state* new_state(form* n){
  return new_form(AND,0,new_form(TRUE,0,0,0),n);
}

void delete_xstate(xstate* xs){
  list* tmp=xs->subs;
  while(tmp!=NULL){
    list* tmp2=tmp;
    tmp=tmp->next;
    free(tmp2);
  }
  free(xs);
}



state* addlabel(int prop,int value,state* s){
  if(s->left->type!=TRUE){
    if(value)
      return new_form(AND,0,new_form(AND,0,s->left,new_form(ATOM,prop,0,0)),s->right);
    else
      return new_form(AND,0,new_form(AND,0,s->left,new_form(NOT,prop,0,0)),s->right);
  } else {
    if(value)
      return new_form(AND,0,new_form(ATOM,prop,0,0),s->right);
    else
      return new_form(AND,0,new_form(NOT,prop,0,0),s->right);
  }
}

void print_xcover(cover* c){
  list* tmp=c;
  while(tmp!=NULL){
    xstate* xs=tmp->el;
    print_form(xs->s);
    tmp=tmp->next;
  }
}

void print_cover(cover* c){
  list* tmp=c;
  while(tmp!=NULL){
    print_form(tmp->el);
    tmp=tmp->next;
  }
}



cover* semantic_branching(int prop,state* s,cover* c,list* tmp){
  form* f=assign(prop,0,s->right);
  if(arequal(f,s->right))
    return c;
  else{
    state* s2;
    f=propagate(f);
    if(f->type!=FALSE){
      s2=addlabel(prop,0,s);
      s2=new_form(AND,0,s2->left,f);
      tmp->el=s2;
      f=assign(prop,1,s->right);
      f=propagate(f);
      if(f->type!=FALSE){
	s2=addlabel(prop,1,s);
	s2=new_form(AND,0,s2->left,f);
	return new_list(s2,c);
      } else 
	return c;
    } else {
      f=assign(prop,1,s->right);
      f=propagate(f);
      s2=addlabel(prop,1,s);
      s2=new_form(AND,0,s2->left,f);
      tmp->el=s2;
      return c;
    }
  }
      
}


int sinF(state* s,int F){
  if(!idimply(s,faircond[F]->promise) || idimply(s,faircond[F]->fulfill))
    return 1;
  else
    return 0;
}

int xsinF(xstate* xs,int F){
  list* tmp=xs->subs;
  while(tmp!=NULL){
    state* sub=tmp->el;
    if(!idimply(sub,faircond[F]->promise))
      tmp=tmp->next;
    else break;
  }
  if(tmp==NULL)
    return 1;
  tmp=xs->subs;
  while(tmp!=NULL){
    state* sub=tmp->el;
    if(idimply(sub,faircond[F]->fulfill))
      tmp=tmp->next;
    else break;
  }
  if(tmp==NULL)
    return 1;
  else
    return 0;
}


int checkcorrectness(xstate* xs){
  int i;
  list* tmp;
  for(i=0;i<fair_number;i++)
    if(!xsinF(xs,i)){
      tmp=xs->subs;
      while(tmp!=NULL){
	state* sub=tmp->el;
	if(sinF(sub,i))
	  return 0;
	else
	  tmp=tmp->next;
      }
    }
  return 1;
}


cover* trypostponement(state* s,cover* c){
  stack* st=new_stack();
  list* ltmp;
  form* tmp;
  xstate *xs=mymalloc(sizeof(xstate));
  form* next=s->right;
  if(verbosity>0){
    printf("state:\n");
    print_form(s);
  }
  next=DNF(s->right);
  next=simplify(next);
  if(next->type==FALSE || next->type==TRUE)
    next=new_form(NEXT,0,next,0);
  if(verbosity>0){
    printf("nextpart after DNF:\n");
    print_form(next);
  }
  next=Xfactoring(next,AND);
  if(verbosity>0){
    printf("nextpart after Xfactoring out of AND:\n");
    print_form(next);
  }
  xs->s=new_form(AND,0,s->left,Xfactoring(next,OR));
  if(verbosity>0){
    printf("nextpart after Xfactoring out of OR:\n");
    print_form(xs->s->right);
  }
  xs->subs=NULL;
  tmp=next;
  while(tmp->type==OR){
    st=push(tmp,st);
    tmp=tmp->left;
  }
  while(tmp!=NULL){
    form* tmp2;
    xs->subs=new_list(new_form(AND,0,s->left,tmp),xs->subs);
    st=pop((void**)&tmp2,st);
    while(tmp2!=NULL && tmp2->right==tmp){
      tmp=tmp2;
      st=pop((void**)&tmp2,st);
    }
    if(tmp2!=NULL){
      st=push(tmp2,st);
      tmp=tmp2->right;
      while(tmp->type==OR){
	st=push(tmp,st);
	tmp=tmp->left;
      }
    } else 
      tmp=tmp2;
  }

  if(checkcorrectness(xs))
    return new_list(xs,c);
  else {
    ltmp=xs->subs;
    while(ltmp!=c){
      xstate* xs2=mymalloc(sizeof(xstate));
      xs2->s=ltmp->el;
      xs2->subs=new_list(ltmp->el,NULL);
      ltmp->el=xs2;
      if(ltmp->next==NULL)
	ltmp->next=c;
      ltmp=ltmp->next;
    }
    ltmp=xs->subs;
    free(xs);
    return ltmp;
  }
  
	      
}


cover* optimize(cover* c){//it merges states with the same next
  list *tmp1,*tmp2,*tmp3;
  tmp1=c;
  while(tmp1!=NULL){
    xstate* xs1=tmp1->el;
    tmp2=tmp1->next;
    tmp3=tmp1;
    while(tmp2!=NULL){
      xstate* xs2=tmp2->el;
      if(idcmp(xs1->s->right->id,xs2->s->right->id)){
	int F;
	int goon=1;
	for(F=0;F<fair_number && goon;F++){
	  int f1,f2;
	  list* tmp4=xs1->subs;
	  while(tmp4!=NULL){
	    state* sub=tmp4->el;
	    if(idimply(sub,faircond[F]->promise))
	      tmp4=tmp4->next;
	    else break;
	  }
	  f1=(tmp4==NULL);
	  tmp4=xs2->subs;
	  while(tmp4!=NULL){
	    state* sub=tmp4->el;
	    if(idimply(sub,faircond[F]->promise))
	      tmp4=tmp4->next;
	    else break;
	  }
	  f2=(tmp4==NULL);
	  if((f1&&!f2) || (!f1&&f2))
	    goon=0;
	  if(goon){
	    tmp4=xs1->subs;
	    while(tmp4!=NULL){
	      state* sub=tmp4->el;
	      if(idimply(sub,faircond[F]->fulfill))
		tmp4=tmp4->next;
	      else break;
	    }
	    f1=(tmp4==NULL);
	    tmp4=xs2->subs;
	    while(tmp4!=NULL){
	      state* sub=tmp4->el;
	      if(idimply(sub,faircond[F]->fulfill))
		tmp4=tmp4->next;
	      else break;
	    }
	    f2=(tmp4==NULL);
	    if((f1&&!f2) || (!f1&&f2))
	      goon=0;
	  }
	}
	if(goon){//merging
	  list* tmp4=xs1->subs;
	  xs1->s=new_form(AND,0,simplify(new_form(OR,0,xs1->s->left,xs2->s->left)),xs1->s->right);
	  while(tmp4->next!=NULL){
	    tmp4=tmp4->next;
	  }
	  tmp4->next=xs2->subs;
	  free(xs2);
	  tmp3->next=tmp2->next;
	  free(tmp2);
	  tmp2=tmp3->next;
	} else {
	  tmp3=tmp3->next;
	  tmp2=tmp2->next;
	}
      } else {
	tmp3=tmp3->next;
	tmp2=tmp2->next;
      }
    }
    tmp1=tmp1->next;
  }
  return c;
}
	    




cover* new_cover(form* f){
  int i;
  cover *c,*c2;
  list* tmp=covered_forms;
  f=simplify(f);
  while(tmp!=NULL){
    form *f2=(form*)tmp->el;
    if(idcmp(f2->id,f->id))
      return f2->c;
    else
      tmp=tmp->next;
  }
  if(verbosity>0){
    printf("formula in input:\n");
    print_form(f);
  }
  f=expand(f);
  if(verbosity>0){
    printf("formula expanded:\n");
    print_form(f);
  }
  f=simplify(f);
  if(verbosity>0){
    printf("formula simplified:\n");
    print_form(f);
  }
  if(f->type==FALSE)
    return NULL;
  c=new_list(new_state(f),NULL);
  for(i=0;i<prop_number;i++){
    tmp=c;
    while(tmp!=NULL){
      c=semantic_branching(i,tmp->el,c,tmp);
      tmp=tmp->next;
    }
      
  }
  if(verbosity>0){
    printf("cover after semantic branching on labels:\n");
    print_cover(c);
  }
  tmp=c;
  c2=NULL;
  while(tmp!=NULL){
    state* s=tmp->el;
    list* tmp2=tmp;
    if(s->right->type==FALSE || s->right->type==TRUE){
      state* s2=new_form(AND,0,s->left,new_form(NEXT,0,s->right,0));
      xstate* xs=mymalloc(sizeof(xstate));
      xs->s=s2;
      xs->subs=new_list(s2,NULL);
      c2=new_list(xs,c2);
    } else
      c2=trypostponement(s,c2);
    tmp=tmp->next;
    myfree(tmp2);
  }
  if(optim){
    if(verbosity>0){
      printf("Cover before optimization:\n");
      print_xcover(c2);
    }
    c2=optimize(c2);
  }
  f->c=c2;
  covered_forms=new_list(f,covered_forms);
  if(verbosity>0){
    printf("Final cover:\n");
    print_xcover(c2);
  }
  return c2;
}
    
