/*
 *  Copyright (C) 2003 2004
 *  Stefano Tonetta <stonetta@dit.unitn.it>
 *  Roberto Sebastiani <rseba@dit.unitn.it>
 *
*/
#include "tree.h"
#include "utils.h"
#include "modella.y.tab.h"


int simply(tree* t1,tree* t2){
  if(treecmp(t1,t2))
    return 1;
  if(t2->type==TRUE)
    return 1;
  if(t1->type==FALSE)
    return 1;
  if(t2->type==AND && (simply(t1,t2->left) && simply(t1,t2->right)))
    return 1;
  if(t1->type==AND && (simply(t1->left,t2) || simply(t1->right,t2)))
    return 1;
  if(t2->type==OR && (simply(t1,t2->left) || simply(t1,t2->right)))
    return 1;
  if(t1->type==OR && (simply(t1->left,t2) && simply(t1->right,t2)))
    return 1;
  if(t2->type==UNTIL && simply(t1,t2->right))
    return 1;
  if(t1->type==RELEASE && simply(t1->right,t2))
    return 1;
  if(t1->type==UNTIL && (simply(t1->left,t2) && simply(t1->right,t2)))
    return 1;
  if(t2->type==RELEASE && (simply(t1,t2->left) && simply(t1,t2->right)))
    return 1;
  if(t1->type==UNTIL && t2->type==UNTIL && (simply(t1->left,t2->left) && simply(t1->right,t2->right)))
    return 1;
  if(t1->type==RELEASE && t2->type==RELEASE && (simply(t1->left,t2->left) && simply(t1->right,t2->right)))
    return 1;
  return 0;
}

tree* rewrite1(tree* t){
  tree* tmp;
  tree* d;
  if(t==NULL)
    return t;
  t->left=rewrite1(t->left);
  t->right=rewrite1(t->right);
  switch(t->type){
  case TRUE:
  case FALSE:
  case ATOM:
  case NOT:
    break;
  case NEXT:
    if(t->left->type==TRUE && t->left->type==FALSE){
      tmp=t;
      t=t->left;
      myfree(t);
    } else if(t->left->type==UNTIL && t->left->left->type==TRUE && t->left->right->type==RELEASE && t->left->right->left->type==FALSE){
      tmp=t;
      t=t->left;
      myfree(tmp);
    } else if(t->left->type==RELEASE && t->left->left->type==FALSE && t->left->right->type==UNTIL && t->left->right->left->type==TRUE){
      tmp=t;
      t=t->left;
      myfree(tmp);
    } else if(t->left->type==OR && t->left->right->type==RELEASE && t->left->right->left->type==FALSE && t->left->right->right->type==UNTIL && t->left->right->right->left->type==TRUE){
      tmp=t;
      t=rewrite1(new_tree(OR,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } else if(t->left->type==OR && t->left->right->type==UNTIL && t->left->right->left->type==TRUE && t->left->right->right->type==RELEASE && t->left->right->right->left->type==FALSE){
      tmp=t;
      t=rewrite1(new_tree(OR,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } else if(t->left->type==AND && t->left->right->type==UNTIL && t->left->right->left->type==TRUE && t->left->right->right->type==RELEASE && t->left->right->right->left->type==FALSE){
      tmp=t;
      t=rewrite1(new_tree(AND,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } else if(t->left->type==AND && t->left->right->type==RELEASE && t->left->right->left->type==FALSE && t->left->right->right->type==UNTIL && t->left->right->right->left->type==TRUE){
      tmp=t;
      t=rewrite1(new_tree(AND,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } 
    break;
  case AND:
    if(simply(t->left,t->right)){
      tmp=t;
      t=t->left;
      delete_tree(tmp->right);
      myfree(tmp);
    } else if(simply(t->right,t->left)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else {
      d=dual(t->right);
      if(simply(t->left,d)){
	tmp=t;
	t=new_tree(FALSE,0,0,0);
	delete_tree(tmp);
      } else {
	delete_tree(d);
	d=dual(t->left);
	if(simply(t->right,d)){
	  tmp=t;
	  t=new_tree(FALSE,0,0,0);
	  delete_tree(tmp);
	} else if(t->left->type==RELEASE && t->right->type==RELEASE && treecmp(t->left->left,t->right->left)){
	  tmp=t;
	  t=rewrite1(new_tree(RELEASE,0,t->left->left,new_tree(AND,0,t->left->right,t->right->right)));
	  delete_tree(tmp->right->left);
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==UNTIL && t->right->type==UNTIL && treecmp(t->left->right,t->right->right)){
	  tmp=t;
	  t=rewrite1(new_tree(UNTIL,0,new_tree(AND,0,t->left->left,t->right->left),t->left->right));
	  delete_tree(tmp->right->right);
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==NEXT && t->right->type==NEXT){
	  tmp=t;
	  t=rewrite1(new_tree(NEXT,0,new_tree(AND,0,t->left->left,t->right->left),0));
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==UNTIL && t->left->left->type==TRUE && t->left->right->type==RELEASE && t->left->right->left->type==FALSE && t->right->type==UNTIL && t->right->left->type==TRUE && t->right->right->type==RELEASE && t->right->right->left->type==FALSE){
	  tmp=t;
	  t=rewrite1(new_tree(UNTIL,0,t->left->left,new_tree(RELEASE,0,t->left->right->left,new_tree(AND,0,t->left->right->right,t->right->right->right))));
	  myfree(tmp->left->right);
	  myfree(tmp->left);
	  myfree(tmp->right->left);
	  myfree(tmp->right->right->left);
	  myfree(tmp->right->right);
	  myfree(tmp->right);
	  myfree(tmp);
	}
      }
      delete_tree(d);
    }
    break;
  case OR:
    if(simply(t->left,t->right)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else if(simply(t->right,t->left)){
      tmp=t;
      t=t->left;
      delete_tree(tmp->right);
      myfree(tmp);
    } else {
      d=dual(t->right);
      if(simply(d,t->left)){
	tmp=t;
	t=new_tree(TRUE,0,0,0);
	delete_tree(tmp);
      } else {
	delete_tree(d);
	d=dual(t->left);
	if(simply(d,t->right)){
	  tmp=t;
	  t=new_tree(TRUE,0,0,0);
	  delete_tree(tmp);
	} else if(t->left->type==UNTIL && t->right->type==UNTIL && treecmp(t->left->left,t->right->left)){
	  tmp=t;
	  t=rewrite1(new_tree(UNTIL,0,t->left->left,new_tree(OR,0,t->left->right,t->right->right)));
	  myfree(tmp->left);
	  delete_tree(tmp->right->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==RELEASE && t->right->type==RELEASE && treecmp(t->left->right,t->right->right)){
	  tmp=t;
	  t=rewrite1(new_tree(RELEASE,0,new_tree(OR,0,t->left->left,t->right->left),t->left->right));
	  myfree(tmp->left);
	  delete_tree(tmp->right->right);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==NEXT && t->right->type==NEXT){
	  tmp=t;
	  t=rewrite1(new_tree(NEXT,0,new_tree(OR,0,t->left->left,t->right->left),0));
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==RELEASE && t->left->left->type==FALSE && t->left->right->type==UNTIL && t->left->right->left->type==TRUE && t->right->type==RELEASE && t->right->left->type==FALSE && t->right->right->type==UNTIL && t->right->right->left->type==TRUE){
	  tmp=t;
	  t=rewrite1(new_tree(RELEASE,0,t->left->left,new_tree(UNTIL,0,t->left->right->left,new_tree(OR,0,t->left->right->right,t->right->right->right))));
	  myfree(tmp->left->right);
	  myfree(tmp->left);
	  myfree(tmp->right->left);
	  myfree(tmp->right->right->left);
	  myfree(tmp->right->right);
	  myfree(tmp->right);
	  myfree(tmp);
	}
      }
      delete_tree(d);
    }
    break;
  case UNTIL:
    if(t->left->type==NEXT && t->right->type==NEXT){
      tmp=t;
      t=rewrite1(new_tree(NEXT,0,new_tree(UNTIL,0,t->left->left,t->right->left),0));
      myfree(tmp->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(t->right->type==FALSE){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else if(simply(t->left,t->right)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else {
      d=dual(t->right);
      if(t->left->type!=TRUE && simply(d,t->left)){
	tmp=t;
	t=rewrite1(new_tree(UNTIL,0,new_tree(TRUE,0,0,0),t->right));
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==NEXT){
	tmp=t;
	t=rewrite1(new_tree(NEXT,0,new_tree(UNTIL,0,t->left,t->right->left),0));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->right->type==UNTIL && simply(t->left,t->right->left)){
	tmp=t;
	t=t->right;
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==RELEASE && t->right->left->type==FALSE && t->right->right->type==UNTIL && t->right->right->left->type==TRUE){
	tmp=t;
	t=t->right;
	myfree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==AND && t->right->right->type==RELEASE && t->right->right->left->type==FALSE && t->right->right->right->type==UNTIL && t->right->right->right->left->type==TRUE){
	tmp=t;
	t=rewrite1(new_tree(AND,0,new_tree(UNTIL,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==AND && t->right->right->type==UNTIL && t->right->right->left->type==TRUE && t->right->right->right->type==RELEASE && t->right->right->right->left->type==FALSE){
	tmp=t;
	t=rewrite1(new_tree(AND,0,new_tree(UNTIL,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      }
      delete_tree(d);
    }
    break;
  case RELEASE:
    if(t->left->type==NEXT && t->right->type==NEXT){
      tmp=t;
      t=rewrite1(new_tree(NEXT,0,new_tree(RELEASE,0,t->left->left,t->right->left),0));
      myfree(tmp->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(t->right->type==TRUE){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else if(simply(t->right,t->left)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else {
      d=dual(t->left);
      if(t->left->type!=FALSE && simply(t->right,d)){
	tmp=t;
	t=rewrite1(new_tree(RELEASE,0,new_tree(FALSE,0,0,0),t->right));
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==NEXT){
	tmp=t;
	t=rewrite1(new_tree(NEXT,0,new_tree(RELEASE,0,t->left,t->right->left),0));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->right->type==RELEASE && simply(t->right->left,t->left)){
	tmp=t;
	t=t->right;
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==UNTIL && t->right->left->type==TRUE && t->right->right->type==RELEASE && t->right->right->left->type==FALSE){
	tmp=t;
	t=t->right;
	myfree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==OR && t->right->right->type==RELEASE && t->right->right->left->type==FALSE && t->right->right->right->type==UNTIL && t->right->right->right->left->type==TRUE){
	tmp=t;
	t=rewrite1(new_tree(OR,0,new_tree(RELEASE,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==OR && t->right->right->type==UNTIL && t->right->right->left->type==TRUE && t->right->right->right->type==RELEASE && t->right->right->right->left->type==FALSE){
	tmp=t;
	t=rewrite1(new_tree(OR,0,new_tree(RELEASE,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      }
      delete_tree(d);
    }
    break;
  }
  return t;
  
}



int pef(tree* t){
  switch(t->type){
  case TRUE:
    return 1;
  case FALSE:
    return 1;
  case ATOM:
    return 0;
  case NOT:
    return 0;
  case NEXT:
    return pef(t->left);
  case AND:
    return pef(t->left) && pef(t->right);
  case OR:
    return pef(t->left) && pef(t->right);
  case UNTIL:
    if(t->left->type==TRUE)
      return 1;
    else
      return pef(t->left) && pef(t->right);
  case RELEASE:
    return pef(t->left) && pef(t->right);
  }
  fatal_error("pef: end reached");
}


int puf(tree* t){
  switch(t->type){
  case TRUE:
    return 1;
  case FALSE:
    return 1;
  case ATOM:
    return 0;
  case NOT:
    return 0;
  case NEXT:
    return puf(t->left);
  case AND:
    return puf(t->left) && puf(t->right);
  case OR:
    return puf(t->left) && puf(t->right);
  case UNTIL:
    return puf(t->left) && puf(t->right);
  case RELEASE:
    if(t->left->type==FALSE)
      return 1;
    else
      return puf(t->left) && puf(t->right);
  }
  
}


tree* rewrite2(tree* t){
  tree* tmp;
  tree* d;
  if(t==NULL)
    return t;
  t->left=rewrite2(t->left);
  t->right=rewrite2(t->right);
  switch(t->type){
  case TRUE:
  case FALSE:
  case ATOM:
  case NOT:
    break;
  case NEXT:
    break;
  case AND:
    if(t->left->type==RELEASE && t->right->type==RELEASE && treecmp(t->left->left,t->right->left)){
      tmp=t;
      t=rewrite2(new_tree(RELEASE,0,t->left->left,new_tree(AND,0,t->left->right,t->right->right)));
      delete_tree(tmp->right->left);
      myfree(tmp->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(t->left->type==UNTIL && t->right->type==UNTIL && treecmp(t->left->right,t->right->right)){
      tmp=t;
      t=rewrite2(new_tree(UNTIL,0,new_tree(AND,0,t->left->left,t->right->left),t->left->right));
      delete_tree(tmp->right->right);
      myfree(tmp->left);
      myfree(tmp->right);
      myfree(tmp);
    } 
    break;
  case OR:
    if(t->left->type==UNTIL && t->right->type==UNTIL && treecmp(t->left->left,t->right->left)){
      tmp=t;
      t=rewrite2(new_tree(UNTIL,0,t->left->left,new_tree(OR,0,t->left->right,t->right->right)));
      myfree(tmp->left);
      delete_tree(tmp->right->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(t->left->type==RELEASE && t->right->type==RELEASE && treecmp(t->left->right,t->right->right)){
      tmp=t;
      t=rewrite2(new_tree(RELEASE,0,new_tree(OR,0,t->left->left,t->right->left),t->left->right));
      myfree(tmp->left);
      delete_tree(tmp->right->right);
      myfree(tmp->right);
      myfree(tmp);
    }
    break;
  case UNTIL:
    if(t->left->type==TRUE && t->right->type==UNTIL){
      tmp=t;
      t=rewrite2(new_tree(UNTIL,0,t->left,t->right->right));
      delete_tree(tmp->right->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(pef(t->right)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    }
    break;
  case RELEASE:
    if(t->left->type==FALSE && t->right->type==RELEASE){
      tmp=t;
      t=rewrite2(new_tree(RELEASE,0,t->left,t->right->right));
      delete_tree(tmp->right->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(puf(t->right)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    }
    break;
  }
  return t;
  
}

  
  
tree* rewrite12(tree* t){
  tree* tmp;
  tree* d;
  if(t==NULL)
    return t;
  t->left=rewrite12(t->left);
  t->right=rewrite12(t->right);
  switch(t->type){
  case TRUE:
  case FALSE:
  case ATOM:
  case NOT:
    break;
  case NEXT:
    if(t->left->type==TRUE && t->left->type==FALSE){
      tmp=t;
      t=t->left;
      myfree(t);
    } else if(t->left->type==UNTIL && t->left->left->type==TRUE && t->left->right->type==RELEASE && t->left->right->left->type==FALSE){
      tmp=t;
      t=t->left;
      myfree(tmp);
    } else if(t->left->type==RELEASE && t->left->left->type==FALSE && t->left->right->type==UNTIL && t->left->right->left->type==TRUE){
      tmp=t;
      t=t->left;
      myfree(tmp);
    } else if(t->left->type==OR && t->left->right->type==RELEASE && t->left->right->left->type==FALSE && t->left->right->right->type==UNTIL && t->left->right->right->left->type==TRUE){
      tmp=t;
      t=rewrite12(new_tree(OR,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } else if(t->left->type==OR && t->left->right->type==UNTIL && t->left->right->left->type==TRUE && t->left->right->right->type==RELEASE && t->left->right->right->left->type==FALSE){
      tmp=t;
      t=rewrite12(new_tree(OR,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } else if(t->left->type==AND && t->left->right->type==UNTIL && t->left->right->left->type==TRUE && t->left->right->right->type==RELEASE && t->left->right->right->left->type==FALSE){
      tmp=t;
      t=rewrite12(new_tree(AND,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } else if(t->left->type==AND && t->left->right->type==RELEASE && t->left->right->left->type==FALSE && t->left->right->right->type==UNTIL && t->left->right->right->left->type==TRUE){
      tmp=t;
      t=rewrite12(new_tree(AND,0,new_tree(NEXT,0,t->left->left,0),t->left->right));
      myfree(tmp);
    } 
    break;
  case AND:
    if(simply(t->left,t->right)){
      tmp=t;
      t=t->left;
      delete_tree(tmp->right);
      myfree(tmp);
    } else if(simply(t->right,t->left)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else {
      d=dual(t->right);
      if(simply(t->left,d)){
	tmp=t;
	t=new_tree(FALSE,0,0,0);
	delete_tree(tmp);
      } else {
	delete_tree(d);
	d=dual(t->left);
	if(simply(t->right,d)){
	  tmp=t;
	  t=new_tree(FALSE,0,0,0);
	  delete_tree(tmp);
	} else if(t->left->type==RELEASE && t->right->type==RELEASE && treecmp(t->left->left,t->right->left)){
	  tmp=t;
	  t=rewrite12(new_tree(RELEASE,0,t->left->left,new_tree(AND,0,t->left->right,t->right->right)));
	  delete_tree(tmp->right->left);
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==UNTIL && t->right->type==UNTIL && treecmp(t->left->right,t->right->right)){
	  tmp=t;
	  t=rewrite12(new_tree(UNTIL,0,new_tree(AND,0,t->left->left,t->right->left),t->left->right));
	  delete_tree(tmp->right->right);
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==NEXT && t->right->type==NEXT){
	  tmp=t;
	  t=rewrite12(new_tree(NEXT,0,new_tree(AND,0,t->left->left,t->right->left),0));
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==UNTIL && t->left->left->type==TRUE && t->left->right->type==RELEASE && t->left->right->left->type==FALSE && t->right->type==UNTIL && t->right->left->type==TRUE && t->right->right->type==RELEASE && t->right->right->left->type==FALSE){
	  tmp=t;
	  t=rewrite12(new_tree(UNTIL,0,t->left->left,new_tree(RELEASE,0,t->left->right->left,new_tree(AND,0,t->left->right->right,t->right->right->right))));
	  myfree(tmp->left->right);
	  myfree(tmp->left);
	  myfree(tmp->right->left);
	  myfree(tmp->right->right->left);
	  myfree(tmp->right->right);
	  myfree(tmp->right);
	  myfree(tmp);
	}
      }
      delete_tree(d);
    }
    break;
  case OR:
    if(simply(t->left,t->right)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else if(simply(t->right,t->left)){
      tmp=t;
      t=t->left;
      delete_tree(tmp->right);
      myfree(tmp);
    } else {
      d=dual(t->right);
      if(simply(d,t->left)){
	tmp=t;
	t=new_tree(TRUE,0,0,0);
	delete_tree(tmp);
      } else {
	delete_tree(d);
	d=dual(t->left);
	if(simply(d,t->right)){
	  tmp=t;
	  t=new_tree(TRUE,0,0,0);
	  delete_tree(tmp);
	} else if(t->left->type==UNTIL && t->right->type==UNTIL && treecmp(t->left->left,t->right->left)){
	  tmp=t;
	  t=rewrite12(new_tree(UNTIL,0,t->left->left,new_tree(OR,0,t->left->right,t->right->right)));
	  myfree(tmp->left);
	  delete_tree(tmp->right->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==RELEASE && t->right->type==RELEASE && treecmp(t->left->right,t->right->right)){
	  tmp=t;
	  t=rewrite12(new_tree(RELEASE,0,new_tree(OR,0,t->left->left,t->right->left),t->left->right));
	  myfree(tmp->left);
	  delete_tree(tmp->right->right);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==NEXT && t->right->type==NEXT){
	  tmp=t;
	  t=rewrite12(new_tree(NEXT,0,new_tree(OR,0,t->left->left,t->right->left),0));
	  myfree(tmp->left);
	  myfree(tmp->right);
	  myfree(tmp);
	} else if(t->left->type==RELEASE && t->left->left->type==FALSE && t->left->right->type==UNTIL && t->left->right->left->type==TRUE && t->right->type==RELEASE && t->right->left->type==FALSE && t->right->right->type==UNTIL && t->right->right->left->type==TRUE){
	  tmp=t;
	  t=rewrite12(new_tree(RELEASE,0,t->left->left,new_tree(UNTIL,0,t->left->right->left,new_tree(OR,0,t->left->right->right,t->right->right->right))));
	  myfree(tmp->left->right);
	  myfree(tmp->left);
	  myfree(tmp->right->left);
	  myfree(tmp->right->right->left);
	  myfree(tmp->right->right);
	  myfree(tmp->right);
	  myfree(tmp);
	}
      }
      delete_tree(d);
    }
    break;
  case UNTIL:
    if(t->left->type==NEXT && t->right->type==NEXT){
      tmp=t;
      t=rewrite12(new_tree(NEXT,0,new_tree(UNTIL,0,t->left->left,t->right->left),0));
      myfree(tmp->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(t->right->type==FALSE){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else if(simply(t->left,t->right)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else {
      d=dual(t->right);
      if(t->left->type!=TRUE && simply(d,t->left)){
	tmp=t;
	t=rewrite12(new_tree(UNTIL,0,new_tree(TRUE,0,0,0),t->right));
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==NEXT){
	tmp=t;
	t=rewrite12(new_tree(NEXT,0,new_tree(UNTIL,0,t->left,t->right->left),0));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->right->type==UNTIL && simply(t->left,t->right->left)){
	tmp=t;
	t=t->right;
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==RELEASE && t->right->left->type==FALSE && t->right->right->type==UNTIL && t->right->right->left->type==TRUE){
	tmp=t;
	t=t->right;
	myfree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==AND && t->right->right->type==RELEASE && t->right->right->left->type==FALSE && t->right->right->right->type==UNTIL && t->right->right->right->left->type==TRUE){
	tmp=t;
	t=rewrite12(new_tree(AND,0,new_tree(UNTIL,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==AND && t->right->right->type==UNTIL && t->right->right->left->type==TRUE && t->right->right->right->type==RELEASE && t->right->right->right->left->type==FALSE){
	tmp=t;
	t=rewrite12(new_tree(AND,0,new_tree(UNTIL,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->left->type==TRUE && t->right->type==UNTIL){
	tmp=t;
	t=rewrite12(new_tree(UNTIL,0,t->left,t->right->right));
	delete_tree(tmp->right->left);
	myfree(tmp->right);
	myfree(tmp);
      } else if(pef(t->right)){
	tmp=t;
	t=t->right;
	delete_tree(tmp->left);
	myfree(tmp);
      }
      delete_tree(d);
    }
    break;
  case RELEASE:
    if(t->left->type==NEXT && t->right->type==NEXT){
      tmp=t;
      t=rewrite12(new_tree(NEXT,0,new_tree(RELEASE,0,t->left->left,t->right->left),0));
      myfree(tmp->left);
      myfree(tmp->right);
      myfree(tmp);
    } else if(t->right->type==TRUE){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else if(simply(t->right,t->left)){
      tmp=t;
      t=t->right;
      delete_tree(tmp->left);
      myfree(tmp);
    } else {
      d=dual(t->left);
      if(t->left->type!=FALSE && simply(t->right,d)){
	tmp=t;
	t=rewrite12(new_tree(RELEASE,0,new_tree(FALSE,0,0,0),t->right));
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==NEXT){
	tmp=t;
	t=rewrite12(new_tree(NEXT,0,new_tree(RELEASE,0,t->left,t->right->left),0));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->right->type==RELEASE && simply(t->right->left,t->left)){
	tmp=t;
	t=t->right;
	delete_tree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==UNTIL && t->right->left->type==TRUE && t->right->right->type==RELEASE && t->right->right->left->type==FALSE){
	tmp=t;
	t=t->right;
	myfree(tmp->left);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==OR && t->right->right->type==RELEASE && t->right->right->left->type==FALSE && t->right->right->right->type==UNTIL && t->right->right->right->left->type==TRUE){
	tmp=t;
	t=rewrite12(new_tree(OR,0,new_tree(RELEASE,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==OR && t->right->right->type==UNTIL && t->right->right->left->type==TRUE && t->right->right->right->type==RELEASE && t->right->right->right->left->type==FALSE){
	tmp=t;
	t=rewrite12(new_tree(OR,0,new_tree(RELEASE,0,t->left,t->right->left),t->right->right));
	myfree(tmp->right);
	myfree(tmp);
      } else if(t->left->type==FALSE && t->right->type==RELEASE){
	tmp=t;
	t=rewrite12(new_tree(RELEASE,0,t->left,t->right->right));
	delete_tree(tmp->right->left);
	myfree(tmp->right);
	myfree(tmp);
      } else if(puf(t->right)){
	tmp=t;
	t=t->right;
	delete_tree(tmp->left);
	myfree(tmp);
      }
      delete_tree(d);
    }
    break;
  }
  return t;
  
}  
