%{
/**YaccFile********************************************************************

  FileName    [modella.y]

  Package     []

  Synopsis    [Yacc input for MoDeLLa input language parser.]

  SeeAlso     [modella.lex]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.9]

******************************************************************************/

#include"tree.h"
#include"main.h"

  int yylex();
  int yyerror(char *s);

  char* getname(tree* t);

%}


%union{ 
  char* prop;
  struct tree* formula; 
}


%start ROOT 
%left<formula>  UNTIL RELEASE 
%nonassoc <formula> FUTURE GLOBALLY
%left <formula> IMPLY EQUIV
%left <formula> OR XOR
%left <formula> AND
%nonassoc <formula> NOT NEXT 
%left IN
%left UNION
%left '=' NE '<' '>' LE GE
%left MOD
%left '+' '-'
%left '*' '/'

%token <formula> FALSE TRUE SFALSE STRUE NTRUE NFALSE '(' ')'
%token <prop> ATOM

%token LBTT SPIN NUSMV

%type <formula> Lbtt_Form Spin_Form NuSMV_Form ListNuSMV_Forms

%%

ROOT:    LBTT Lbtt_Form { parsing_tree=$2; }
|        SPIN Spin_Form { parsing_tree=$2; }
|        NUSMV NuSMV_Form { parsing_tree=$2; }
;
Lbtt_Form: AND Lbtt_Form Lbtt_Form { $$=new_tree(AND,0,$2,$3); }
|        OR Lbtt_Form Lbtt_Form { $$=new_tree(OR,0,$2,$3); }
|        UNTIL Lbtt_Form Lbtt_Form { $$=new_tree(UNTIL,0,$2,$3); }
|        RELEASE Lbtt_Form Lbtt_Form { $$=new_tree(RELEASE,0,$2,$3); }
|        FUTURE Lbtt_Form { $$=new_tree(UNTIL,0,new_tree(TRUE,0,NULL,NULL),$2); }
|        GLOBALLY Lbtt_Form { $$=new_tree(RELEASE,0,new_tree(FALSE,0,NULL,NULL),$2); }
|        NEXT Lbtt_Form { $$=new_tree(NEXT,0,$2,NULL); }
|        FALSE { $$=new_tree(FALSE,0,NULL,NULL); }
|        TRUE { $$=new_tree(TRUE,0,NULL,NULL); }
|        NOT Lbtt_Form { $$=new_tree(NOT,0,$2,NULL); }
|        ATOM { $$=new_tree(ATOM,$1,NULL,NULL); }
;
Spin_Form: Spin_Form AND Spin_Form { $$=new_tree(AND,0,$1,$3); }
|          Spin_Form OR Spin_Form { $$=new_tree(OR,0,$1,$3); }
|          Spin_Form UNTIL Spin_Form { $$=new_tree(UNTIL,0,$1,$3); }
|          Spin_Form RELEASE Spin_Form { $$=new_tree(RELEASE,0,$1,$3); }
|          Spin_Form IMPLY Spin_Form { $$=new_tree(OR,0,new_tree(NOT,0,$1,0),$3); }
|          Spin_Form EQUIV Spin_Form { $$=new_tree(AND,0,new_tree(OR,0,new_tree(NOT,0,$1,0),$3),new_tree(OR,0,new_tree(NOT,0,dup_tree($3),0),dup_tree($1))); }
|          FUTURE Spin_Form { $$=new_tree(UNTIL,0,new_tree(TRUE,0,NULL,NULL),$2); }
|          GLOBALLY Spin_Form { $$=new_tree(RELEASE,0,new_tree(FALSE,0,NULL,NULL),$2); }
|          FALSE { $$=new_tree(FALSE,0,NULL,NULL); }
|          TRUE { $$=new_tree(TRUE,0,NULL,NULL); }
|          NEXT Spin_Form { $$=new_tree(NEXT,0,$2,NULL); }
|          NOT Spin_Form { $$=new_tree(NOT,0,$2,NULL); }
|          '(' Spin_Form ')' { $$=$2; }
|          ATOM { $$=new_tree(ATOM,$1,NULL,NULL); }
;
NuSMV_Form: NuSMV_Form AND NuSMV_Form { $$=new_tree(AND,0,$1,$3); }
|          NuSMV_Form OR NuSMV_Form { $$=new_tree(OR,0,$1,$3); }
|          NuSMV_Form UNTIL NuSMV_Form { $$=new_tree(UNTIL,0,$1,$3); }
|          NuSMV_Form RELEASE NuSMV_Form { $$=new_tree(RELEASE,0,$1,$3); }
|          NuSMV_Form IMPLY NuSMV_Form { $$=new_tree(OR,0,new_tree(NOT,0,$1,0),$3); }
|          NuSMV_Form EQUIV NuSMV_Form { $$=new_tree(AND,0,new_tree(OR,0,new_tree(NOT,0,$1,0),$3),new_tree(OR,0,new_tree(NOT,0,dup_tree($3),0),dup_tree($1))); }
|          FUTURE NuSMV_Form { $$=new_tree(UNTIL,0,new_tree(TRUE,0,NULL,NULL),$2); }
|          GLOBALLY NuSMV_Form { $$=new_tree(RELEASE,0,new_tree(FALSE,0,NULL,NULL),$2); }
|          FALSE { $$=new_tree(FALSE,0,NULL,NULL); }
|          TRUE { $$=new_tree(TRUE,0,NULL,NULL); }
|          NEXT NuSMV_Form { $$=new_tree(NEXT,0,$2,NULL); }
|          NOT NuSMV_Form { $$=new_tree(NOT,0,$2,NULL); }
|          '(' NuSMV_Form ')' { $$=$2; }
|          ATOM { $$=new_tree(ATOM,$1,NULL,NULL); }
|          NuSMV_Form '=' NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s = %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form NE NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+9)*sizeof(char));
  sprintf(name,"( %s != %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form '<' NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s < %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form '>' NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s > %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form LE NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+9)*sizeof(char));
  sprintf(name,"( %s <= %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form GE NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+9)*sizeof(char));
  sprintf(name,"( %s >= %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form '+' NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s + %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form '-' NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s - %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form '*' NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s * %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form '/' NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s / %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form MOD NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+8)*sizeof(char));
  sprintf(name,"( %s mod %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          '{' ListNuSMV_Forms '}' { 
  char *name = (char*)mymalloc((strlen($2->name)+5)*sizeof(char));
  sprintf(name,"{ %s }",$2->name);
  $2->name=name;
  $$=$2; 
}
|          NuSMV_Form IN NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+9)*sizeof(char));
  sprintf(name,"( %s in %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
|          NuSMV_Form UNION NuSMV_Form {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+12)*sizeof(char));
  sprintf(name,"( %s union %s )",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
}
;
ListNuSMV_Forms: NuSMV_Form { $$=$1; }
|          NuSMV_Form ',' ListNuSMV_Forms {
  char *namel=getname($1);
  char *namer=getname($3);
  char *name = (char*)mymalloc((strlen(namel)+strlen(namer)+2)*sizeof(char));
  sprintf(name,"%s,%s",namel,namer);
  myfree(namel);
  myfree(namer);
  delete_tree($1);
  delete_tree($3);
  $$=new_tree(ATOM,name,NULL,NULL);
;
}

%%

char* getname(tree* t){
  char *name;
  if(t==NULL){
    fprintf(stderr,"fatal error in getname: NULL pointer found\n");
    exit(-1);
  } else {
    switch(t->type){
    case TRUE:
      name=(char*)mymalloc(5*sizeof(char));
      sprintf(name,"TRUE");
      return name;
    case FALSE:
      name=(char*)mymalloc(6*sizeof(char));
      sprintf(name,"FALSE");
      return name;
    case ATOM:
      name=(char*)mymalloc((strlen(t->name)+1)*sizeof(char));
      sprintf(name,"%s",t->name);
      return name;
    case NOT:
      name=(char*)mymalloc((strlen(t->name)+2)*sizeof(char));
      sprintf(name,"!%s",t->left->name);
      return name;
    }
  }
  fprintf(stderr,"fatal error in getname: end reached\n");
  exit(-1);
}
    
