/**CFile***********************************************************************

  FileName    [struct.c]

  Package     [utils]

  Synopsis    [Useful abstract types.]

  Description [This file provides the function to manipulate stacks an
  queues.]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.8.2]

******************************************************************************/

#include <stdio.h>
#include "struct.h"
#include "utils.h"

list* new_list(void* el,list* next){
  list* l=(list*)mymalloc(sizeof(list));
  l->el=el;
  l->next=next;
  return l;
}

void delete_list(list* l){
  myfree(l);
}

stack* new_stack(){
  return NULL;
}

int empty_stack(stack* s){
  return (s == NULL);
}

stack* pop(void** el,stack* s){
  *el=NULL;
  if(!empty_stack(s)){
    list* tmp=s;
    *el=s->el;
    s=s->next;
    delete_list(tmp);
  } 
  return s;
}

stack* push(void* el,stack* s){
  return new_list(el,s);
}

queue* new_queue(){
  queue *q=(queue*)mymalloc(sizeof(queue));
  q->head=NULL;
  return q;
}

int empty_queue(queue* q){
  return (q->head == NULL);
}

void* dequeue(queue* q){
  void* ret=NULL;
  if(q->head!=NULL){
    list* tmp=q->head;
    ret=q->head->el;
    q->head=q->head->next;
    delete_list(tmp);
  }
  return ret;
}

void enqueue(void *el,queue*q){
  list* l=new_list(el,NULL);
  if(empty_queue(q))
    q->head=l;
  else 
    q->tail->next=l;
  q->tail=l;
}
