/**CFile***********************************************************************

  FileName    [utils.c]

  Package     [utils]

  Synopsis    [Useful functions]

  Description [This file provides useful functions.]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.8.2]

******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <stdlib.h>

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [It prints a message and exit]

  Description        [The function is called when something bad
  happened. It prints an error message and exits.]

******************************************************************************/
void fatal_error(char *message){
  fprintf(stderr,"fatal error: %s\n",message);
  exit(77);
}

/**Function********************************************************************

  Synopsis           [It returns a power of 2]

  Description        [Wrapper of the function pow(2,x). We first check
  if x is ]

******************************************************************************/
size_t modella_pow2(size_t x){
  double dret = pow(2.0,(double)x);
  size_t iret = (size_t)dret;
  if ( errno == ERANGE || iret==0 ) {
    fprintf(stderr,"too much memory required\n");
    exit(-1);
  } 
  return iret;
}



/**Function*******************************************************************/
void* mymalloc(size_t size){
  void* p=(void*)malloc(size);
  if(p==NULL){
    fprintf(stderr,"full memory\n");
    exit(-1);
  } else
    return p;
}
    

/**Function*******************************************************************/
void myfree(void *p){
  if(p==NULL){
    fprintf(stderr,"try to free null pointer\n" );
    exit(-1);
  } else
    free(p);
}
	  

/**Function*******************************************************************/
unsigned int modella_log2(unsigned int n){
  unsigned int i=0;
  n=n/2;
  while(n!=0){
    n=n/2;
    i++;
  }
  return i;
}
